/* 
 * The MIT License (MIT)
 *
 * Copyright (c) 2019 Ha Thach (tinyusb.org)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

#include "bsp/board.h"
#include "tusb.h"

#if CFG_TUD_MSC

#define MSC_RO_ORIGIN (0x10180000)
#define DISK_BLOCK_NUM 1024
#define DISK_BLOCK_SIZE 512

extern const unsigned char disk0_img[];

struct disk_lba {
    uint8_t data[DISK_BLOCK_SIZE];
};


// whether host does safe-eject
static bool ejected = false;

static struct disk_lba lba_cache;
static int lba_cached = -1;

struct drive {
    const uint8_t lun;
    const char vid[8];
    const char pid[16];
    const char rev[4];
    const uint32_t n_lba;
};

#define N_DRIVES 2

static uint8_t drv_active[N_DRIVES] = {1, 0};


const char global_vid[8] = "DLX";
const char host_pid[16] = "MEP Tools";
const char vault_pid[16] = "MEP Vault";
const char global_rev[4] = "0.1";

static const struct drive Drives[N_DRIVES] = { 
    { 0, "DLX", "MEP Tools", "0.1", 1024},
    { 1, "DLX", "MEP Vault", "0.1", 1024}
};


uint8_t tud_msc_get_maxlun_cb(void)
{
  return 1;
}

// Invoked when received SCSI_CMD_INQUIRY
// Application fill vendor id, product id and revision with string up to 8, 16, 4 characters respectively
void tud_msc_inquiry_cb(uint8_t lun, uint8_t vendor_id[8], uint8_t product_id[16], uint8_t product_rev[4])
{
    const struct drive *drv;
    if (lun >= N_DRIVES) {
        return;
    }
    drv = &Drives[lun];
    memcpy(vendor_id  , drv->vid, strlen(drv->vid));
    memcpy(product_id , drv->pid, strlen(drv->pid));
    memcpy(product_rev, drv->rev, strlen(drv->rev));
}

// Invoked when received Test Unit Ready command.
// return true allowing host to read/write this LUN e.g SD card inserted
bool tud_msc_test_unit_ready_cb(uint8_t lun)
{
    if (lun >= N_DRIVES) {
        tud_msc_set_sense(lun, SCSI_SENSE_NOT_READY, 0x3a, 0x00);
        return false;
    }
    if (drv_active[lun]) {
        return true;
    } else {
        tud_msc_set_sense(lun, SCSI_SENSE_NOT_READY, 0x3a, 0x00);
        return false;
    }
}

// Invoked when received SCSI_CMD_READ_CAPACITY_10 and SCSI_CMD_READ_FORMAT_CAPACITY to determine the disk size
// Application update block count and block size
void tud_msc_capacity_cb(uint8_t lun, uint32_t* block_count, uint16_t* block_size)
{
    if (lun >= N_DRIVES) {
        *block_count = DISK_BLOCK_NUM;
        *block_size  = DISK_BLOCK_SIZE;
    } else {
        *block_count = Drives[lun].n_lba;
        *block_size  = DISK_BLOCK_SIZE;
    }
}

// Invoked when received Start Stop Unit command
// - Start = 0 : stopped power mode, if load_eject = 1 : unload disk storage
// - Start = 1 : active mode, if load_eject = 1 : load disk storage
bool tud_msc_start_stop_cb(uint8_t lun, uint8_t power_condition, bool start, bool load_eject)
{
    const struct drive *drv;
    (void) power_condition;

    if (lun < N_DRIVES) {
        drv = &Drives[lun];
        if ( load_eject )
        {
            if (start)
            {
                drv_active[lun] = 1;
            }else
            {
                drv_active[lun] = 0;
            }
        }
    }
    return true;
}

// Callback invoked when received READ10 command.
// Copy disk's data to buffer (up to bufsize) and return number of copied bytes.
int32_t tud_msc_read10_cb(uint8_t lun, uint32_t lba, uint32_t offset, void* buffer, uint32_t bufsize)
{

    const struct drive *drv;
    uint8_t const *addr;
    if (lun >= N_DRIVES)
        return -1;
    drv = &Drives[lun];

    if ( lba >= drv->n_lba )
        return -1;




    if (lun == 0) {
        struct disk_lba *msc_ro = (struct disk_lba *)disk0_img;
        memcpy(buffer, msc_ro[lba].data + offset, bufsize);
        asm volatile("DMB");
        return bufsize;
    } else {
#if 0
        if ( lba != lba_cached) {
            if (lba == 0) {
                memcpy(lba_cache.data, mbr.data, sizeof(struct disk_lba)); 
                lba_cached = lba;
                return 0;
            }
            memcpy(lba_cache.data, msc_ro[lba].data, sizeof(struct disk_lba)); 
            lba_cached = lba;
            return 0;
        }
        memcpy(buffer, lba_cache.data + offset, bufsize);
#endif
        memset(buffer, 0xA5, bufsize);
        return bufsize;
    }
}

bool tud_msc_is_writable_cb (uint8_t lun)
{
    if (lun != 1) {
      /* Read only. */
      return false;
    } else {
        return true;
    }
}

// Callback invoked when received WRITE10 command.
// Process data in buffer to disk's storage and return number of written bytes
int32_t tud_msc_write10_cb(uint8_t lun, uint32_t lba, uint32_t offset, uint8_t* buffer, uint32_t bufsize)
{
  (void) lun;
  // out of ramdisk
  if ( lba >= DISK_BLOCK_NUM )
      return -1;
  /* Read only. Silently fail. */
  (void) lba; (void) offset; (void) buffer;
  return bufsize;
}

// Callback invoked when received an SCSI command not in built-in list below
// - READ_CAPACITY10, READ_FORMAT_CAPACITY, INQUIRY, MODE_SENSE6, REQUEST_SENSE
// - READ10 and WRITE10 has their own callbacks
int32_t tud_msc_scsi_cb (uint8_t lun, uint8_t const scsi_cmd[16], void* buffer, uint16_t bufsize)
{
    // read10 & write10 has their own callback and MUST not be handled here

    void const* response = NULL;
    int32_t resplen = 0;

    // most scsi handled is input
    bool in_xfer = true;

    switch (scsi_cmd[0])
    {
        case SCSI_CMD_PREVENT_ALLOW_MEDIUM_REMOVAL:
            // Host is about to read/write etc ... better not to disconnect disk
            resplen = 0;
            break;

        default:
            // Set Sense = Invalid Command Operation
            tud_msc_set_sense(lun, SCSI_SENSE_ILLEGAL_REQUEST, 0x20, 0x00);

            // negative means error -> tinyusb could stall and/or response with failed status
            resplen = -1;
            break;
    }

    // return resplen must not larger than bufsize
    if ( resplen > bufsize ) resplen = bufsize;

    if ( response && (resplen > 0) )
    {
        if(in_xfer)
        {
            memcpy(buffer, response, resplen);
        }else
        {
            // SCSI output
        }
    }

    return resplen;
}

#endif
